unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, VLBasicGenericFilter, VLGenericFilter,
  Mitov.Types, LPComponent, SLCommonFilter, VLCommonFilter,
  VLConnectedComponents, System.ImageList, Vcl.ImgList, Vcl.ComCtrls,
  Vcl.StdCtrls, ILBasicClassifier, ILNaiveBayes, ILRadialBasisFunctionNetwork;

type
  TForm1 = class(TForm)
    Label1: TLabel;
    Label2: TLabel;
    Label4: TLabel;
    Button1: TButton;
    TrainingListView: TListView;
    TestListView: TListView;
    TrainingImageList: TImageList;
    TestImageList: TImageList;
    ILRadialBasisFunctionNetwork1: TILRadialBasisFunctionNetwork;
    procedure FormCreate(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure ILRadialBasisFunctionNetwork1Result(ASender: TObject;
      AFeatures: ISLRealBuffer; AResult: TILRadialBasisFunctionResult);
  private
    FTrainingLabels   : ISLRealBuffer;
    FTestLabels       : ISLRealBuffer;
    FResult           : Integer;

  private
    procedure PopulateImageList( AImageList : TImageList; AFileName : String );
    function  LoadLabels( AFileName : String ) : ISLRealBuffer;
    procedure PopulateListViewLabels( AListView : TListView; ALabels : ISLRealBuffer );
    procedure PopulateListView( AListView : TListView );
    function  ExtractImageFeatures( ABmp : TBitmap ) : ISLRealBuffer;
    function  ExtractAllFeatures( AImageList : TImageList ) : ISLRealBufferArray;

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  Mitov.Containers.List;

const
  IMAGE_WIDTH = 28;
  IMAGE_HEIGHT = 28;

//---------------------------------------------------------------------------
procedure TForm1.PopulateImageList( AImageList : TImageList; AFileName : String );
var
  ALines  : IStringArrayList;
  ABitmap : TBitmap;
  I       : Integer;
  X       : Integer;
  APixels : PColor;

begin
  ALines := TStringArrayList.Create();
  ABitmap := TBitmap.Create();
  ABitmap.Width := IMAGE_WIDTH;
  ABitmap.Height := IMAGE_HEIGHT;

  ALines.LoadFromFile( AFileName );

  ABitmap.PixelFormat := pf32bit;
  APixels := ABitmap.ScanLine[ 0 ];
  Inc( APixels, IMAGE_WIDTH );

  for I := 0 to ALines.Count - 1 do
    begin
    if((( I mod IMAGE_HEIGHT ) = 0 ) and ( I > 0 )) then
      begin
      AImageList.AddMasked( ABitmap, clRed );
      APixels := ABitmap.ScanLine[ 0 ];
      Inc( APixels, IMAGE_WIDTH );
      end;

    for X := 28 downto 1 do
      begin
      Dec( APixels );
      case( ALines[ I ][ X ] ) of
        '+' :
          APixels^ := clGray;

        '#' :
          APixels^ := clBlack;

        else
          APixels^ := clWhite;

        end;
      end;
    end;

  AImageList.AddMasked( ABitmap, clRed );
  ABitmap.Free();
end;
//---------------------------------------------------------------------------
function TForm1.LoadLabels( AFileName : String ) : ISLRealBuffer;
var
  ALines  : IStringArrayList;
  I       : Integer;

begin
  ALines := TStringArrayList.Create();
  ALines.LoadFromFile( AFileName );

  Result := TSLRealBuffer.CreateSize( ALines.Count );

  for I := 0 to ALines.Count - 1 do
    Result[ I ] := ALines[ I ].ToInteger();

end;
//---------------------------------------------------------------------------
procedure TForm1.PopulateListViewLabels( AListView : TListView; ALabels : ISLRealBuffer );
var
  I : Integer;

begin
  AListView.Items.BeginUpdate();
  for I := 0 to AListView.Items.Count - 1 do
    AListView.Items[ I ].Caption := Round( ALabels[ I ] ).ToString();

  AListView.Items.EndUpdate();
end;
//---------------------------------------------------------------------------
procedure TForm1.PopulateListView( AListView : TListView );
var
  I     : Integer;
  AItem : TListItem;

begin
  AListView.Items.BeginUpdate();
  for I := 0 to AListView.LargeImages.Count - 1 do
    begin
    AItem := AListView.Items.Add();
    AItem.ImageIndex := I;
    end;

  AListView.Items.EndUpdate();
end;
//---------------------------------------------------------------------------
function TForm1.ExtractImageFeatures( ABmp : TBitmap ) : ISLRealBuffer;
var
  I       : Integer;
  APixels : PColor;
  APtr    : PReal;

begin
  Result := TSLRealBuffer.CreateSize( ABmp.Width * ABmp.Height );
  APtr := Result.Write();
  ABmp.PixelFormat := pf32bit;
  APixels := ABmp.ScanLine[ ABmp.Height - 1 ];
  for I := 0 to ABmp.Width * ABmp.Height - 1 do
    begin
    if( APixels^ <> clWhite ) then
      APtr^ := 1

    else
      APtr^ := 0;

    Inc( APixels );
    Inc( APtr );
    end;

end;
//---------------------------------------------------------------------------
function TForm1.ExtractAllFeatures( AImageList : TImageList ) : ISLRealBufferArray;
var
  ABitmap : TBitmap;
  I       : Cardinal;
  AResult : TSLRealBufferList;

begin
  AResult := TSLRealBufferList.Create();
  ABitmap := TBitmap.Create();
  for I := 0 to AImageList.Count - 1 do
    begin
    AImageList.GetBitmap( I, ABitmap );
    AResult.Append( ExtractImageFeatures( ABitmap ));
    end;

  ABitmap.Free();

  Result := AResult;
end;
//---------------------------------------------------------------------------
procedure TForm1.Button1Click(Sender: TObject);
var
  ATrainingData : ISLRealBufferArray;
  ATestFeatures : ISLRealBufferArray;

  I             : Integer;
  AErrorsCount  : Integer;

begin
  ATrainingData := ExtractAllFeatures( TrainingImageList );

  for I := 0 to ATrainingData.Count - 1 do
    ILRadialBasisFunctionNetwork1.Train( ATrainingData[ I ], FTrainingLabels[ I ] );

  TestListView.Items.BeginUpdate();

  ATestFeatures := ExtractAllFeatures( TestImageList );
  AErrorsCount := 0;
  for I := 0 to TestImageList.Count - 1 do
    begin
    ILRadialBasisFunctionNetwork1.Predict( ATestFeatures[ I ] );

    if( FResult <> FTestLabels[ I ] ) then
      begin
      TestListView.Items[ I ].Caption := 'Error ' + FResult.ToString() + '(' + Round( FTestLabels[ I ] ).ToString() + ')';
      Inc( AErrorsCount );
      end

    else
      TestListView.Items[ I ].Caption := FResult.ToString() + '(' + Round( FTestLabels[ I ] ).ToString() + ')';

    end;

  TestListView.Items.EndUpdate();

  Label1.Caption := AErrorsCount.ToString() + ' Errors of ' + TestImageList.Count.ToString() + ' - ' + ( ( TestImageList.Count - AErrorsCount ) * 100 / TestImageList.Count ).ToString() + '% correct';
end;
//---------------------------------------------------------------------------
procedure TForm1.FormCreate(Sender: TObject);
begin
  PopulateImageList( TrainingImageList, 'C:\Data\OCR\trainingimages' );
  FTrainingLabels := LoadLabels( 'C:\Data\OCR\traininglabels' );

  PopulateListView( TrainingListView );
  PopulateListViewLabels( TrainingListView, FTrainingLabels );

  PopulateImageList( TestImageList, 'C:\Data\OCR\testimages' );
  FTestLabels := LoadLabels( 'C:\Data\OCR\testlabels' );

  PopulateListView( TestListView );
  PopulateListViewLabels( TestListView, FTestLabels );
end;
//---------------------------------------------------------------------------
procedure TForm1.ILRadialBasisFunctionNetwork1Result(ASender: TObject; AFeatures: ISLRealBuffer; AResult: TILRadialBasisFunctionResult);
begin
  if( AResult.Count = 0 ) then
    begin
    FResult := -1;
    Exit;
    end;

  FResult := Round( AResult[ 0 ].Neuron.Category );
end;
//---------------------------------------------------------------------------
end.
